from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from pydantic_xml import BaseXmlModel, element, wrapped

from py_altium365.base.connection_handler import ConnectionHandler
from py_altium365.base.enums import PrtGlobalService
from py_altium365.connection.soapy_con import (
    SoapHeader,
    SoapMethod,
    SoapResponse,
    SoapyCon,
)

if TYPE_CHECKING:
    from py_altium365.altium_api import AltiumApi


class SoapHeaderApi(
    SoapHeader,
    tag="Header",
    ns="soap",
):
    """SOAP header for an API call."""

    api_version: str = element(tag="APIVersion", default="2.0", nsmap={"": "http://tempuri.org/"})


class SoapMethodLogin(
    SoapMethod,
    tag="Login",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP method for login."""

    username: str = element(tag="Username")
    password: str = element(tag="Password")


class SoapProfilePicture(BaseXmlModel, tag="ProfilePicture"):
    """SOAP Profile picture."""

    small_link: Optional[str] = element(tag="Small", default=None)
    medium_link: Optional[str] = element(tag="Medium", default=None)
    large_link: Optional[str] = element(tag="Large", default=None)
    full_link: Optional[str] = element(tag="Full", default=None)


class SoapParameter(BaseXmlModel, tag="Parameter", ns="temp", nsmap={"temp": "http://tempuri.org/"}):
    """SOAP Parameter."""

    name: str = element(tag="Name")
    value: Optional[str] = element(tag="Value", default=None)


class SoapLoginResult(
    BaseXmlModel,
    tag="LoginResult",
    ns="temp",
    nsmap={"i": "http://www.w3.org/2001/XMLSchema-instance", "temp": "http://tempuri.org/"},
):
    """SOAP Login result for the SOAP login call."""

    message: Optional[str] = element(tag="Message", default=None)
    success: bool = element(tag="Success")
    session_handle: Optional[str] = element(tag="SessionHandle", default=None)
    user_name: Optional[str] = element(tag="UserName", default=None)
    first_name: Optional[str] = element(tag="FirstName", default=None)
    last_name: Optional[str] = element(tag="LastName", default=None)
    full_name: Optional[str] = element(tag="FullName", default=None)
    time_zone_sid_key: Optional[str] = element(tag="TimeZoneSidKey", default=None)
    language_locale_key: Optional[str] = element(tag="LanguageLocaleKey", default=None)
    locale_sid_key: Optional[str] = element(tag="LocaleSidKey", default=None)
    user_rights_string: Optional[str] = element(tag="UserRightsString", default=None)
    contact_guid: Optional[str] = element(tag="ContactGUID", default=None)
    email: Optional[str] = element(tag="Email", default=None)
    last_login_date: Optional[datetime] = element(tag="LastLoginDate", default=None)
    password_expired: Optional[bool] = element(tag="PasswordExpired", default=None)
    profile_picture: Optional[SoapProfilePicture] = element(tag="ProfilePicture", default=None)
    parameters: List[SoapParameter] = wrapped(
        path="Parameters",
        tag="Parameter",
        default=[],
    )
    fault_code: Optional[str] = element(tag="FaultCode", default=None)
    allowed_features: List[str] = wrapped(
        path="AllowedFeatures",
        entity=element(tag="string", ns="a", default=None),
        nsmap={
            "a": "http://schemas.microsoft.com/2003/10/Serialization/Arrays",
        },
        default=[],
    )
    secured_session_handle: Optional[str] = element(tag="SecuredSessionHandle", default=None)


class SoapLoginResponse(
    SoapResponse,
    tag="LoginResponse",
    ns="temp",
    nsmap={"temp": "http://tempuri.org/"},
):
    """SOAP Login response."""

    login_result: SoapLoginResult


class SoapMethodGetPrtGlobalServiceUrl(
    SoapMethod,
    tag="GetPRT_GlobalServiceUrl",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP method for getting a PRT global service URL."""

    handle: Optional[str] = element(tag="Handle", default=None)
    service_name: str = element(tag="ServiceName")
    set_name: str = element(tag="SetName")


class SoapGetPrtGlobalServiceUrlResponse(
    SoapResponse,
    tag="GetPRT_GlobalServiceUrlResponse",
    ns="temp",
    nsmap={"temp": "http://tempuri.org/"},
):
    """SOAP response for getting a PRT global service URL."""

    service_url: Optional[str] = element(tag="ServiceURL", nsmap={"temp": "http://tempuri.org/"}, default=None)


class SoapyConPortal(SoapyCon):
    """SOAP connection to the Altium portal."""

    def __init__(self, altium_api: AltiumApi) -> None:
        """
        Initialize the SOAP connection to the Altium portal.
        :param altium_api: The Altium API object.
        """
        super().__init__(ConnectionHandler.get_instance(), "https://portal365.altium.com/?cls=soap")
        self._altium_api = altium_api

    def login_user(self, username: str, password: str) -> SoapLoginResult:
        """
        Login a user to the Altium portal.
        :param username: The altium username.
        :param password: The altium password.
        :return: The user login result.
        """
        response = self._send_command(
            SoapHeaderApi(),
            SoapMethodLogin(username=username, password=password),
            return_method=SoapLoginResponse,
        )

        return response.login_result

    def get_prt_global_service_url(self, service: PrtGlobalService, handle: Optional[str] = None) -> Optional[str]:
        """
        Get a PRT global service URL.
        :param service: The PRT global service to get the URL for.
        :param handle: A handle for the service.
        :return: The service URL.
        """
        if handle is None and service == service.is_guid_req:  # type: ignore
            raise ValueError(f"Handle is required for {service.name} service")
        if not service.is_guid_req:
            handle = None

        response = self._send_command(
            SoapHeaderApi(),
            SoapMethodGetPrtGlobalServiceUrl(
                handle=handle,
                service_name=service.service_name,
                set_name=service.set_name,
            ),
            return_method=SoapGetPrtGlobalServiceUrlResponse,
        )
        return response.service_url



from typing import Optional

from py_altium365.connection.json_con_search_async import JsonConSearchAsync
from py_altium365.connection.soapy_con_service_discovery import SoapyConServiceDiscovery
from py_altium365.connection.vault.soapy_con_vault import SoapConVault
from py_altium365.connection.vault.soapy_con_vault_base import AluItem, SoapMethodOption


class AltiumApiWorkspace:
    """Altium API workspace class"""

    def __init__(self, workspace_url: str, service_discovery: SoapyConServiceDiscovery):
        """
        Initialize the Altium API workspace object
        :param workspace_url: The URL of the workspace
        :param service_discovery: The service discovery object
        """

        if service_discovery.user_info is None:
            raise ConnectionError("Failed to get user info")
        self.workspace_url: str = workspace_url
        self._service_discovery: SoapyConServiceDiscovery = service_discovery
        self.session_guid: str = service_discovery.user_info.session_id
        if self._service_discovery.service_urls.SEARCHBASE is None:
            raise ConnectionError("Failed to get search base URL")
        self._vault = SoapConVault(self)

    def create_search_object(self) -> JsonConSearchAsync:
        """
        Create a search object
        :return:
        """
        if self._service_discovery.service_urls.SEARCHBASE is None:
            raise ConnectionError("Failed to get search base URL")
        return JsonConSearchAsync(self, self._service_discovery.service_urls.SEARCHBASE, self.session_guid, self.workspace_url.strip(":443").strip("https://"))

    def get_item_from_guid(self, guid: str) -> Optional[AluItem]:
        """
        Get an item from the vault using its GUID
        :param guid: The GUID of the item to retrieve
        :return: A list of AluItem objects matching the GUID
        """
        items = self._vault.get_alu_items(options=[SoapMethodOption.INCLUDE_ALL_CHILD_OBJECTS], p_filter="GUID='" + guid + "'")
        return items[0] if len(items) > 0 else None



import typing
from typing import Generic, Optional, TypeVar

from pydantic_xml import BaseXmlModel, element
from requests import Session


class SoapHeader(
    BaseXmlModel,
    tag="Header",
    ns="soap",
    nsmap={"soap": "http://schemas.xmlsoap.org/soap/envelope/"},
):
    """Base class for SOAP header."""


class SoapMethod(
    BaseXmlModel,
    tag="Method",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """Base class for SOAP method."""


MethodTypeT = TypeVar("MethodTypeT", bound=SoapMethod)


class SoapBody(
    BaseXmlModel,
    Generic[MethodTypeT],
    tag="Body",
    ns="soap",
):
    """SOAP body."""

    method: MethodTypeT


HeaderTypeT = TypeVar("HeaderTypeT", bound=SoapHeader)
BodyTypeT = TypeVar("BodyTypeT", bound=SoapBody)


class SoapEnvelopeNoHeader(
    BaseXmlModel,
    Generic[BodyTypeT],
    tag="Envelope",
    ns="soap",
    nsmap={
        "soap": "http://schemas.xmlsoap.org/soap/envelope/",
    },
):
    """SOAP envelope."""

    body: BodyTypeT


class SoapEnvelope(
    BaseXmlModel,
    Generic[HeaderTypeT, BodyTypeT],
    tag="Envelope",
    ns="soap",
    nsmap={
        "soap": "http://schemas.xmlsoap.org/soap/envelope/",
    },
):
    """SOAP envelope."""

    header: HeaderTypeT = element(default=None)
    body: BodyTypeT


class SoapResponse(
    SoapMethod,
    tag="Response",
    nsmap={"i": "http://www.w3.org/2001/XMLSchema-instance"},
):
    """Base class for SOAP response."""

    message: Optional[str] = element(
        tag="Message",
        ns="i",
        default=None,
    )


# pylint: disable=too-few-public-methods
class SoapyCon:
    """Base class for SOAP connection."""

    def __init__(self, session: Session, url: str):
        """
        Initialize the SoapyCon object
        :param session: The main connection session
        :param url: The URL to send the SOAP request to
        """
        self._session: Session = session
        self._url: str = url

    ReturnMethodT = TypeVar("ReturnMethodT", bound=SoapMethod)

    @typing.no_type_check
    def _send_command(
        self,
        header: Optional[SoapHeader],
        method: SoapMethod,
        return_method: typing.Type[ReturnMethodT],
        soap_action: Optional[str] = None,
        return_header=SoapHeader,
    ) -> ReturnMethodT:
        """
        Send a SOAP command
        :param header: The SOAP header object
        :param method: The SOAP method object
        :param return_method: The return SOAP method class
        :param soap_action: The optional SOAP action
        :param return_header: The return header class
        :return: The return method data
        """
        if soap_action is None:
            soap_action = method.__xml_tag__

        if header is None:
            request_shape = SoapEnvelopeNoHeader[SoapBody[type(method)]]
        else:
            request_shape = SoapEnvelope[
                type(header),
                SoapBody[type(method)],
            ]

        if return_header is None:
            return_shape = SoapEnvelopeNoHeader[return_method,]
        else:
            return_shape = SoapEnvelope[
                return_header,
                SoapBody[return_method],
            ]

        if header is None:
            envelope = request_shape(
                body=SoapBody(
                    method=method,
                ),
            )
        else:
            envelope = request_shape(
                header=header,
                body=SoapBody(
                    method=method,
                ),
            )

        headers = {
            "content-type": "text/xml",
            "SOAPAction": soap_action,
            "User-Agent": "Altium Designer",
        }

        response = self._session.post(self._url, data=envelope.to_xml(encoding="UTF-8"), headers=headers)

        if response.status_code != 200:
            raise ConnectionError(f'Failed to send "{soap_action}" command!')

        ret = return_shape.from_xml(response.text.encode("utf-8"))
        return ret.body.method



from typing import Optional

from requests import Session


# pylint: disable=too-few-public-methods
class ConnectionHandler:
    """
    This class is a singleton class that returns a session object
    """

    def __init__(self):
        """
        Initialize the ConnectionHandler object
        """
        self.session = Session()

    @staticmethod
    def get_instance() -> Session:
        """
        Get the session object
        :return: The session object
        """
        global _connection_handler_instance  # pylint: disable=global-statement
        if _connection_handler_instance is None:
            _connection_handler_instance = ConnectionHandler()
        return _connection_handler_instance.session

    @staticmethod
    def set_instance(session: Session):
        """
        Set the session object
        :param session: The session object
        """
        global _connection_handler_instance  # pylint: disable=global-statement
        if _connection_handler_instance is None:
            _connection_handler_instance = ConnectionHandler()
        _connection_handler_instance.session = session


_connection_handler_instance: Optional[ConnectionHandler] = None  # pylint: disable=invalid-name



from typing import Tuple, Any

import pytest

from py_altium365.altium_api_workspace import AltiumApiWorkspace


def create_workspace_api(mocker) -> Tuple[AltiumApiWorkspace, Any]:
    workspace_url = "test_workspace_url"
    service_discovery = mocker.Mock()
    service_discovery.user_info.session_id = "test_session_id"
    service_discovery.service_urls.SEARCHBASE = "test_search_base_url"
    return AltiumApiWorkspace(workspace_url, service_discovery), service_discovery


def test_init(mocker):
    workspace_url = "test_workspace_url"
    service_discovery = mocker.Mock()
    service_discovery.user_info.session_id = "test_session_id"
    service_discovery.service_urls.SEARCHBASE = "test_search_base_url"

    api_workspace = AltiumApiWorkspace(workspace_url, service_discovery)
    assert api_workspace.workspace_url == workspace_url
    assert api_workspace._service_discovery == service_discovery
    assert api_workspace.session_guid == service_discovery.user_info.session_id
    assert api_workspace._service_discovery.service_urls.SEARCHBASE == "test_search_base_url"


def test_init_no_user_info(mocker):
    workspace_url = "test_workspace_url"
    service_discovery = mocker.Mock()
    service_discovery.user_info = None

    with pytest.raises(ConnectionError):
        AltiumApiWorkspace(workspace_url, service_discovery)


def test_init_no_search_base_url(mocker):
    workspace_url = "test_workspace_url"
    service_discovery = mocker.Mock()
    service_discovery.user_info.session_id = "test_session_id"
    service_discovery.service_urls.SEARCHBASE = None

    with pytest.raises(ConnectionError):
        AltiumApiWorkspace(workspace_url, service_discovery)


def test_create_search_object(mocker):
    api, service_discovery = create_workspace_api(mocker)
    service_discovery.service_urls.SEARCHBASE = "test_search_base_url"
    mocker.patch("py_altium365.altium_api_workspace.JsonConSearchAsync", return_value="TEST")

    assert api.create_search_object() == "TEST"


def test_create_search_object_no_search_base_url(mocker):
    api, service_discovery = create_workspace_api(mocker)
    service_discovery.service_urls.SEARCHBASE = None

    with pytest.raises(ConnectionError):
        api.create_search_object()



from dataclasses import dataclass
from enum import Enum


@dataclass()
class PrtService:
    """
    Data class for the PRT Service
    """

    service_name: str
    set_name: str
    is_guid_req: bool


class PrtGlobalService(PrtService, Enum):
    """
    Enum for the PRT Global Service
    """

    CIIVA = ("CiivaApi", "Secure", False)
    ENGAGEMENT_PAGE = ("EngagementPage", "Secure", False)
    VAULT_CONTENT_SERVICE = ("VaultContentServiceURL", "Secure", True)
    STAT_SERVICE = ("StatServiceURL", "Secure", True)
    WORKSPACE = ("WorkspacesUrl", "Secure", True)
    PART_CATALOG = ("PartCatalogUrl", "Secure", True)
    A365_USER_PROFILE = ("A365UserProfile", "Secure", True)
    A365_VALIDATESESION = ("A365ValidateSession", "Secure", True)
    AD_PAYMENT_SERVICE = ("ADPaymentService", "Secure", True)
    APP_REGISTRY = ("AppRegistryUrl", "Secure", True)

    # def __new__(cls, name: str, set_name: str, is_guid_req: bool):
    #     obj = object.__new__(cls)
    #     obj._value_ = name
    #     obj.service_name = name  # type: ignore
    #     obj.set_name = set_name  # type: ignore
    #     obj.is_guid_req = is_guid_req  # type: ignore
    #     return obj

    # def __init__(self, name: str, set_name: str, is_guid_req: bool):
    #     self.name = name
    #     self.set_name = set_name
    #     self.is_guid_req = is_guid_req



from typing import List, Optional, Union

from py_altium365.altium_api_workspace import AltiumApiWorkspace
from py_altium365.base.enums import PrtGlobalService
from py_altium365.connection.soapy_con_portal import SoapyConPortal
from py_altium365.connection.soapy_con_service_discovery import SoapyConServiceDiscovery
from py_altium365.connection.soapy_con_workspace import (
    SoapyConWorkspace,
    UserWorkspaceInfo,
)


class AltiumApi:
    """
    Altium API class
    """

    def __init__(self) -> None:
        """
        Initialize the Altium API object
        """

        # Global variables

        self._portal_con: SoapyConPortal = SoapyConPortal(self)
        self._session_guid: Optional[str] = None
        self._service_urls: dict[str, str] = {}
        self._workspace_con: Optional[SoapyConWorkspace] = None

        # Workspace variables
        self._service_discovery_con: Optional[SoapyConServiceDiscovery] = None

    def login(self, username: str, password: str, return_message: bool = False) -> Union[str, bool]:
        """
        Login to the Altium API
        :param username: The altium username
        :param password: The altium password
        :param return_message: If the login fails, return the message
        :return: True if the login was successful, False otherwise or the message if return_message is True
        """
        try:
            user_login = self._portal_con.login_user(username, password)
            if not user_login.success:
                if return_message and user_login.message is not None:
                    return user_login.message
                return False
            self._session_guid = user_login.session_handle
            self._workspace_con = SoapyConWorkspace(self)
        except ConnectionError:
            return False
        return True

    def login_workspace(
        self, workspace: Union[UserWorkspaceInfo, str], username: str, password: str, return_message: bool = False, force_login: bool = False
    ) -> Optional[AltiumApiWorkspace]:
        """
        Login to a workspace
        :param workspace: The url of the workspace or the UserWorkspaceInfo object
        :param username: The altium username
        :param password: The altium password
        :param return_message: If the login fails, return the message
        :param force_login: Force a user login even if the session is already active
        :return: True if the login was successful, False otherwise or the message if return_message is True
        """
        if self._session_guid is None or force_login:
            msg = self.login(username, password, return_message)
            if msg is not True:
                return None
        if isinstance(workspace, UserWorkspaceInfo):
            workspace = workspace.hosting_url
        if not isinstance(workspace, str):
            return None
        service_discovery_con = SoapyConServiceDiscovery(workspace)
        service_discovery_con.login(username, password)
        if not service_discovery_con.user_info:
            return None
        return AltiumApiWorkspace(workspace, service_discovery_con)

    def get_service_url(self, service: PrtGlobalService, force_request: bool = False) -> Optional[str]:
        """
        Get a service URL
        :param service: The service to get the URL for
        :param force_request: If the URL should be requested again from altium or use the cached URL
        :return: The service URL
        """
        if service.name in self._service_urls and not force_request:
            return self._service_urls[service.name]
        url = self._portal_con.get_prt_global_service_url(service, self._session_guid)
        if url is not None:
            self._service_urls[service.name] = url
        return url

    def get_user_workspaces(self) -> List[UserWorkspaceInfo]:
        """
        Get the user workspaces
        :return: The user workspaces
        """
        if self._workspace_con is None:
            return []
        if self._session_guid is None:
            return []
        return self._workspace_con.get_user_workspaces(self._session_guid)



from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, List, Optional

from pydantic_xml import BaseXmlModel, element, wrapped

from py_altium365.base.connection_handler import ConnectionHandler
from py_altium365.connection.soapy_con import SoapyCon

if TYPE_CHECKING:
    from py_altium365.altium_api_workspace import AltiumApiWorkspace


class SoapMethodOption(str, Enum):
    """Enum for SOAP method options."""

    INCLUDE_ALL_CHILD_OBJECTS = "IncludeAllChildObjects=True"


class AluObject(BaseXmlModel, tag="item", nsmap={"temp": "http://tempuri.org/"}, ns="temp"):
    """Base class for Altium Vault objects."""

    guid: Optional[str] = element(tag="GUID", default=None)
    hrid: Optional[str] = element(tag="HRID", default=None)
    created_at: Optional[datetime] = element(tag="CreatedAt", default=None)
    last_modified_at: Optional[datetime] = element(tag="LastModifiedAt", default=None)
    created_by_guid: Optional[str] = element(tag="CreatedByGUID", default=None)
    last_modified_by_guid: Optional[str] = element(tag="LastModifiedByGUID", default=None)
    created_by_name: Optional[str] = element(tag="CreatedByName", default=None)
    last_modified_by_name: Optional[str] = element(tag="LastModifiedByName", default=None)


class AluShareableObject(AluObject):
    """Base class for shareable Altium Vault objects."""

    sharing_control: Optional[int] = element(tag="SharingControl", default=None)
    access_rights: Optional[int] = element(tag="AccessRights", default=None)


class AluItemRevisionParameter(AluObject):
    """Class representing an item revision parameter in Altium Vault."""

    parameter_value: Optional[str] = element(tag="ParameterValue", default=None)
    item_revision_guid: Optional[str] = element(tag="ItemRevisionGUID", default=None)
    parameter_type_guid: Optional[str] = element(tag="ParameterTypeGUID", default=None)
    parameter_real_value: Optional[float] = element(tag="ParameterRealValue", default=None)


class AluLifeCycleStateChange(AluObject):
    """Class representing a life cycle state change in Altium Vault."""

    item_revision_guid: Optional[str] = element(tag="ItemRevisionGUID", default=None)
    life_cycle_state_transition_guid: Optional[str] = element(tag="LifeCycleStateTransitionGUID", default=None)
    life_cycle_state_after_guid: Optional[str] = element(tag="LifeCycleStateAfterGUID", default=None)
    note: Optional[str] = element(tag="Note", default=None)


class AluItemRevision(AluShareableObject, tag="item", nsmap={"temp": "http://tempuri.org/"}, ns="temp"):
    """Class representing an item revision in Altium Vault."""

    revision_id: Optional[str] = element(tag="RevisionId", default=None)
    ancestor_item_revision_guid: Optional[str] = element(tag="AncestorItemRevisionGUID", default=None)
    description: Optional[str] = element(tag="Description", default=None)
    comment: Optional[str] = element(tag="Comment", default=None)
    lifecycle_state_guid: Optional[str] = element(tag="LifeCycleStateGUID", default=None)
    item_guid: Optional[str] = element(tag="ItemGUID", default=None)
    release_date: Optional[datetime] = element(tag="ReleaseDate", default=None)
    item_hrid: Optional[str] = element(tag="ItemHRID", default=None)
    item_description: Optional[str] = element(tag="ItemDescription", default=None)
    content_type_guid: Optional[str] = element(tag="ContentTypeGUID", default=None)
    folder_guid: Optional[str] = element(tag="FolderGUID", default=None)
    revision_id_levels: List[str] = wrapped(path="RevisionIdLevels", entity=element(tag="item", default=None), default=[])
    revision_id_separators: List[str] = wrapped(
        path="RevisionIdSeparators",
        entity=element(tag="item", default=None),
        default=[],
    )
    revision_parameters: List[AluItemRevisionParameter] = wrapped(
        path="RevisionParameters",
        default=[],
    )
    state_changes: List[AluLifeCycleStateChange] = wrapped(
        path="StateChanges",
        default=[],
    )
    is_shared: Optional[bool] = element(tag="IsShared", default=None)
    is_payload_shared: Optional[bool] = element(tag="IsPayloadShared", default=None)
    source_vault_guid: Optional[str] = element(tag="SourceVaultGUID", default=None)
    source_guid: Optional[str] = element(tag="SourceGUID", default=None)
    is_visible: Optional[bool] = element(tag="IsVisible", default=None)
    is_applicable: Optional[bool] = element(tag="IsApplicable", default=None)
    is_active: Optional[bool] = element(tag="IsActive", default=None)


class AluTag(AluObject):
    """Class representing a tag in Altium Vault."""

    tag_family_guid: Optional[str] = element(tag="TagFamilyGUID", default=None)
    parant_tag_guid: Optional[str] = element(tag="ParentTagGUID", default=None)
    sub_tags: List[AluTag] = wrapped(
        path="SubTags",
        default=[],
    )


class AluItemParameter(AluObject):
    """Class representing an item parameter in Altium Vault."""

    parameter_value: Optional[str] = element(tag="ParameterValue", default=None)
    item_guid: Optional[str] = element(tag="ItemGUID", default=None)
    parameter_type_guid: Optional[str] = element(tag="ParameterTypeGUID", default=None)
    parameter_real_value: Optional[float] = element(tag="ParameterRealValue", default=None)


class AluItem(AluShareableObject, tag="item", nsmap={"temp": "http://tempuri.org/"}, ns="temp"):
    """Class representing an item in Altium Vault."""

    description: Optional[str] = element(tag="Description", default=None)
    folder_guid: Optional[str] = element(tag="FolderGUID", default=None)
    lifecycle_definition_guid: Optional[str] = element(tag="LifeCycleDefinitionGUID", default=None)
    revision_naming_scheme_guid: Optional[str] = element(tag="RevisionNamingSchemeGUID", default=None)
    content_type_guid: Optional[str] = element(tag="ContentTypeGUID", default=None)
    revisions: List[AluItemRevision] = wrapped(
        path="Revisions",
        default=[],
    )
    is_shared: bool = element(tag="IsShared", default=False)
    tags: List[AluTag] = wrapped(
        path="Tags",
        default=[],
    )
    item_parameters: List[AluItemParameter] = wrapped(
        path="ItemParameters",
        default=[],
    )
    is_active: bool = element(tag="IsActive", default=False)


class SoapConVaultBase(SoapyCon):
    """Base class for SOAP connection to Altium Vault."""

    def __init__(self, altium_workspace: "AltiumApiWorkspace"):
        """
        Initialize the SOAP connection to Altium Vault.
        :param altium_workspace: The Altium API workspace object.
        """
        super().__init__(ConnectionHandler.get_instance(), altium_workspace.workspace_url + "/vault/?cls=soap")
        self._altium_workspace = altium_workspace



from py_altium365.altium_api import AltiumApi
from py_altium365.base.enums import PrtGlobalService


def test_user_login_correct(mocker):
    api = AltiumApi()

    mock_portal_con = mocker.Mock()
    mock_portal_con.login_user.return_value.success = True
    mock_portal_con.login_user.return_value.session_handle = "test_session_handle"
    api._portal_con = mock_portal_con

    mock_soapy_con_workspace = mocker.patch("py_altium365.altium_api.SoapyConWorkspace")
    mock_soapy_con_workspace.return_value = "test_soapy_con_workspace"

    assert api.login("test_user", "test_pass")
    assert api._session_guid == "test_session_handle"
    assert api._workspace_con == "test_soapy_con_workspace"
    assert mock_portal_con.login_user.called
    assert mock_soapy_con_workspace.called


def test_user_login_incorrect(mocker):
    api = AltiumApi()

    mock_portal_con = mocker.Mock()
    mock_portal_con.login_user.return_value.success = False
    mock_portal_con.login_user.return_value.message = "test_message"
    api._portal_con = mock_portal_con

    assert not api.login("test_user", "test_pass")
    assert api._session_guid is None
    assert api._workspace_con is None
    assert mock_portal_con.login_user.called


def test_user_login_incorrect_with_message(mocker):
    api = AltiumApi()

    mock_portal_con = mocker.Mock()
    mock_portal_con.login_user.return_value.success = False
    mock_portal_con.login_user.return_value.message = "test_message"
    api._portal_con = mock_portal_con

    assert api.login("test_user", "test_pass", return_message=True) == "test_message"
    assert api._session_guid is None
    assert api._workspace_con is None
    assert mock_portal_con.login_user.called


def test_user_login_connection_error(mocker):
    api = AltiumApi()

    mock_portal_con = mocker.Mock()
    mock_portal_con.login_user.side_effect = ConnectionError
    api._portal_con = mock_portal_con

    assert not api.login("test_user", "test_pass")
    assert api._session_guid is None
    assert api._workspace_con is None
    assert mock_portal_con.login_user.called


def test_login_workspace(mocker):
    api = AltiumApi()

    api.login = mocker.Mock()
    api.login.return_value = True

    mock_soapy_con_workspace = mocker.patch("py_altium365.altium_api.SoapyConServiceDiscovery")
    sd_login = mocker.Mock()
    mock_soapy_con_workspace.login = sd_login
    sd_login.return_value.user_info = True
    api._workspace_con = mock_soapy_con_workspace

    mocker.patch("py_altium365.altium_api.AltiumApiWorkspace", return_value="test_workspace")

    assert api.login_workspace("test_workspace", "test_user", "test_pass") == "test_workspace"


def test_get_service_url_cache(mocker):
    api = AltiumApi()

    api._service_urls = {
        PrtGlobalService.WORKSPACE.name: "test_workspace_url",
    }

    mock_portal_con = mocker.Mock()
    mock_portal_con.get_prt_global_service_url.return_value = "test_portal_url"
    api._portal_con = mock_portal_con

    assert api.get_service_url(PrtGlobalService.WORKSPACE) == "test_workspace_url"
    assert api.get_service_url(PrtGlobalService.WORKSPACE, True) == "test_portal_url"
    mock_portal_con.get_prt_global_service_url.assert_called_once_with(PrtGlobalService.WORKSPACE, None)


def test_get_service_url_no_cache(mocker):
    api = AltiumApi()

    api._service_urls = {}

    mock_portal_con = mocker.Mock()
    mock_portal_con.get_prt_global_service_url.return_value = "test_portal_url"
    api._portal_con = mock_portal_con

    assert api.get_service_url(PrtGlobalService.WORKSPACE) == "test_portal_url"
    mock_portal_con.get_prt_global_service_url.assert_called_once_with(PrtGlobalService.WORKSPACE, None)



from __future__ import annotations

import re
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Dict, Generic, List, Optional, Tuple, TypeVar, Union

from pydantic import BaseModel, Field

from py_altium365.base.connection_handler import ConnectionHandler
from py_altium365.connection.json_con import JsonCon, JsonRequest, JsonReturn

if TYPE_CHECKING:
    from py_altium365.altium_api_workspace import AltiumApiWorkspace


class FacedType(str, Enum):
    """Faced type."""

    NO_TYPE = ""
    CAPACITANCE = "0629FA77_2DBB3E_2D4045_2D91F0_2DCC1164D3D0AADD420E8DDD8B445E911A0601BB2B6D53"
    RESISTANCE = "B90F0DAE_2DB695_2D41F5_2DBCB0_2D0DE5F75C9E50DD420E8DDD8B445E911A0601BB2B6D53"
    VOLTAGE = "28379228_2DD94F_2D4F3B_2D8038_2D3FE85C36E292DD420E8DDD8B445E911A0601BB2B6D53"
    TOLERANCE = "935791AE_2D95D8_2D4D5E_2DB810_2DE9B66A56E5A5DD420E8DDD8B445E911A0601BB2B6D53"
    FREQUENCY = "F4CAEC49_2DEE33_2D46C9_2DAF4A_2D260721F7AA26DD420E8DDD8B445E911A0601BB2B6D53"
    DISTANCE = "B0C104C5_2D1A99_2D4DE5_2DA539_2DD5FC817D4B3EDD420E8DDD8B445E911A0601BB2B6D53"
    POWER = "A40E567D_2D1324_2D4823_2D8379_2DFB7E896E18B9DD420E8DDD8B445E911A0601BB2B6D53"

    @classmethod
    def _missing_(cls, value):  # type: ignore
        return FacedType.NO_TYPE


encoding_decoding_naming = {
    "_5F": "_",
    "_20": " ",
    "_28": "(",
    "_29": ")",
    "_2C": ",",
    "_2D": "-",
    "_2E": ".",
    "_2F": "/",
    "_40": "@",
    "_5B": "[",
    "_5D": "]",
    "_5E": "^",
}


class JsonDtoSearchConditionBaseQuery(BaseModel):
    """Base class for search conditions."""


JsonDtoSearchConditionBaseQueryTypeT = TypeVar("JsonDtoSearchConditionBaseQueryTypeT", bound=JsonDtoSearchConditionBaseQuery)


class JsonDtoSearchConditionTerm(BaseModel):
    """Search condition term."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionTerm")
    field: str = Field(alias="Field")
    value: str = Field(alias="Value")


class JsonDtoSearchConditionStrictQuery(JsonDtoSearchConditionBaseQuery):
    """Strict search query."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionStrictQuery")
    term: JsonDtoSearchConditionTerm = Field(alias="Term")


class JsonDtoSearchConditionRangeQuery(JsonDtoSearchConditionBaseQuery):
    """Range search query."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionRangeQuery")
    field: str = Field(alias="Field")
    field_type: int = Field(alias="FieldType", default=1)
    min: float = Field(alias="Min", default=0)
    max: float = Field(alias="Max", default=0)
    precision_step: int = Field(alias="PrecisionStep", default=0)
    min_inclusive: bool = Field(alias="MinInclusive", default=False)
    max_inclusive: bool = Field(alias="MaxInclusive", default=False)


class JsonDtoSearchConditionWildcardQuery(JsonDtoSearchConditionBaseQuery):
    """Wildcard search query."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionWildcardQuery")
    term: JsonDtoSearchConditionTerm = Field(alias="Term")


class JsonDtoSearchConditionBooleanQueryItem(BaseModel, Generic[JsonDtoSearchConditionBaseQueryTypeT]):
    """Boolean search query item."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionBooleanQueryItem")
    item: JsonDtoSearchConditionBaseQueryTypeT = Field(alias="Item")
    occur: int = Field(alias="Occur", default=0)


class JsonDtoSearchConditionBooleanQuery(JsonDtoSearchConditionBaseQuery):
    """Boolean search query."""

    rtype: str = Field(alias="$type", default="DtoSearchConditionBooleanQuery")
    items: List[JsonDtoSearchConditionBooleanQueryItem] = Field(alias="Items", default=[])


class JsonSearchAsyncRequest(JsonRequest):
    """JSON search async request."""

    rtype: str = Field(alias="$type", default="SearchRequest")
    condition: JsonDtoSearchConditionBooleanQuery = Field(alias="Condition")
    sort_fields: List[str] = Field(alias="SortFields", default=[])
    return_fields: List[str] = Field(alias="ReturnFields", default=[])
    start: int = Field(alias="Start", default=0)
    limit: int = Field(alias="Limit", default=0)
    include_facets: bool = Field(alias="IncludeFacets", default=True)
    use_only_best_facets: bool = Field(alias="UseOnlyBestFacets", default=False)
    include_debug_info: bool = Field(alias="IncludeDebugInfo", default=False)
    ingnore_case_field_names: bool = Field(alias="IgnoreCaseFieldNames", default=False)


class JsonField(BaseModel):
    """Field."""

    name: str = Field(alias="Name")
    value: str = Field(alias="Value")
    field_type: int = Field(alias="FieldType", default=0)


class JsonDocument(BaseModel):
    """Document."""

    score: float = Field(alias="Score")
    fields: List[JsonField] = Field(alias="Fields")


class JsonCounter(BaseModel):
    """Counter."""

    value: str = Field(alias="Value")
    count: int = Field(alias="Count")


class JsonFacetedCounter(BaseModel):
    """Faceted counter."""

    faced_name: str = Field(alias="FacetName")
    faced_type: FacedType = Field(default=FacedType.NO_TYPE)
    total_hit_count: int = Field(alias="TotalHitCount")
    counters: List[JsonCounter] = Field(alias="Counters")
    support_range: bool = Field(alias="SupportRange", default=False)


class JsonSearchAsyncReturn(JsonReturn):
    """JSON search async return."""

    documents: List[JsonDocument] = Field(alias="Documents", default=[])
    faceted_counters: List[JsonFacetedCounter] = Field(alias="FacetedCounters", default=[])
    total: int = Field(alias="Total", default=0)
    success: bool = Field(alias="Success", default=False)


class SearchDataBase(BaseModel):
    """Search data base. The base class for search data."""

    altium_workspace: object
    id: str = Field(alias="Id", default="")
    parameters: Dict[str, Union[str, float]] = Field(default_factory=lambda: {}, alias="Parameters")
    hrid: str = Field(alias="HRID", default="")
    created_at: datetime = Field(alias="CreatedAt", default=datetime(1899, 12, 31))
    created_by: str = Field(alias="CreatedBy", default="")
    modified_by: str = Field(alias="ModifiedBy", default="")
    latest_revision: bool = Field(alias="LatestRevision", default=False)
    updated: datetime = Field(alias="Updated", default=datetime(1899, 12, 31))
    created_by_guid: str = Field(alias="CreatedByGUID", default="")
    updated_by_guid: str = Field(alias="UpdatedByGUID", default="")
    app_type: str = Field(alias="AppType", default="")
    url: str = Field(alias="Url", default="")
    submit_date: datetime = Field(alias="SubmitDate", default=datetime(1899, 12, 31))
    language: str = Field(alias="Language", default="")
    folder_guid: str = Field(alias="FolderGUID", default="")
    folder_full_path: str = Field(alias="FolderFullPath", default="")
    description: str = Field(alias="Description", default="")
    naming_scheme_guid: str = Field(alias="NamingSchemeGuid", default="")
    cat: str = Field(alias="Cat", default="")
    content_type_guid: str = Field(alias="ContentTypeGUID", default="")
    text: str = Field(alias="Text", default="")
    life_cycle: str = Field(alias="LifeCycle", default="")
    acl: str = Field(alias="ACL", default="")
    item_hrid: str = Field(alias="ItemHRID", default="")
    comment: str = Field(alias="Comment", default="")
    life_cycle_state_guid: str = Field(alias="LifeCycleStateGUID", default="")
    item_guid: str = Field(alias="ItemGUID", default="")
    revision_id: str = Field(alias="RevisionId", default="")
    ancestor_revision_guid: str = Field(alias="AncestorRevisionGUID", default="")
    item_description: str = Field(alias="ItemDescription", default="")
    release_note: str = Field(alias="ReleaseNote", default="")
    release_date: datetime = Field(alias="ReleaseDate", default=datetime(1899, 12, 31))
    release_data_num: datetime = Field(alias="ReleaseDataNum", default=datetime(1899, 12, 31))
    life_cycle_definition_guid: str = Field(alias="LifeCycleDefinitionGUID", default="")
    update_date: datetime = Field(alias="Update Date", default=datetime(1899, 12, 31))
    content_type: str = Field(alias="ContentType", default="")

    def get_item(self) -> Optional[str]:
        """
        Get the item from the Altium workspace using the item GUID.
        :return:
        """
        if hasattr(self.altium_workspace, "get_item_from_guid"):
            return self.altium_workspace.get_item_from_guid(self.item_guid)
        return None


class SearchDataType(str, Enum):
    """Search data type."""

    NO_TYPE = ""
    COMPONENT = "Component"
    FOOTPRINT = "Footprint"
    SYMBOL = "Symbol"
    DATASHEET = "Datasheet"
    PROJECT_TEMPLATE = "Project Template"
    SCHEMATIC_TEMPLATE = "Schematic Template"
    COMPONENT_TEMPLATE = "Component Template"
    SCIPT = "Script"
    LAYERSTACK = "Layerstack"

    @classmethod
    def _missing_(cls, value):  # type: ignore
        return FacedType.NO_TYPE


NOT_COUNTED_SEARCH_PARAMETERS = [
    JsonFacetedCounter(FacetName="LatestRevision", faced_type=FacedType.NO_TYPE, TotalHitCount=0, Counters=[]),
    JsonFacetedCounter(FacetName="FolderFullPath", faced_type=FacedType.NO_TYPE, TotalHitCount=0, Counters=[]),
]


class JsonConSearchAsync(JsonCon):
    """JSON connection search async."""

    def __init__(self, altium_workspace: "AltiumApiWorkspace", url: str, session_guid: str, host: str):
        """
        Initialize the JsonConSearchAsync object
        :param url: The URL to send the JSON request to
        :param session_guid: The session GUID
        :param host: The host for the host parameter
        """
        super().__init__(ConnectionHandler.get_instance(), url + "/v1.0/searchasync", session_guid, host)
        self._altium_workspace = altium_workspace
        self._counters_up_to_date = False
        self._search_parameters: List[JsonDtoSearchConditionBooleanQueryItem] = []
        self._search_counters: List[JsonFacetedCounter] = []
        self._total_hits: int = 0

        self._update_search_names_and_counters()

    def add_search_parameter(self, name: str, value: Union[str, List[str]], dtype: FacedType = FacedType.NO_TYPE, remove_old: bool = False) -> bool:
        """
        Add a search parameter
        :param name:
        :param value:
        :param dtype:
        :param remove_old:
        :return:
        """
        self._update_search_names_and_counters()
        if isinstance(value, str):
            value = [value]
        counter: Optional[JsonFacetedCounter] = None
        for c in self._search_counters:
            if c.faced_name == name and c.faced_type == dtype:
                counter = c
                break
        if counter is None:
            return False
        full_name = self._get_index_name_from_name_and_type(counter.faced_name, counter.faced_type)
        s_param: Optional[JsonDtoSearchConditionBooleanQueryItem] = self._get_search_parameter(full_name)
        old_values: list[str] = []
        if s_param is not None and not remove_old:
            old_values = self.get_search_parameter(name, dtype)

        for v in old_values:
            if v not in value:
                value.append(v)
        if s_param is not None:
            self._search_parameters.remove(s_param)

        if len(value) == 0:
            return True

        querry_items = []
        occurance = 1 if len(value) > 1 else 0
        for v in value:
            querry_items.append(
                JsonDtoSearchConditionBooleanQueryItem(
                    Item=JsonDtoSearchConditionStrictQuery(Term=JsonDtoSearchConditionTerm(Field=full_name, Value=v)), Occur=occurance
                )
            )
        if len(querry_items) > 1:
            s_param = JsonDtoSearchConditionBooleanQueryItem(Item=JsonDtoSearchConditionBooleanQuery(Items=querry_items), Occur=0)
        else:
            s_param = querry_items[0]

        self._search_parameters.append(s_param)
        self._counters_up_to_date = False
        return True

    def remove_search_parameter(self, name: str, value: Union[str, List[str], None] = None, dtype: FacedType = FacedType.NO_TYPE) -> bool:
        """
        Remove a search parameter
        :param name: The name of the search parameter
        :param value: The specific value to remove, if None all values are removed
        :param dtype: The data type of the search parameter
        :return: If the search parameter was removed
        """
        old_values = []
        if value is not None:
            if isinstance(value, str):
                value = [value]
            old_values = self.get_search_parameter(name, dtype)
            for v in value:
                if v in old_values:
                    old_values.remove(v)
        self.add_search_parameter(name, old_values, dtype, remove_old=True)
        self._counters_up_to_date = False
        return True

    def get_search_parameter(self, name: str, dtype: FacedType = FacedType.NO_TYPE) -> List[str]:
        """
        Get the search parameter
        :param name: The name of the search parameter
        :param dtype: The data type of the search parameter
        :return: A list of the search parameter values
        """
        ret_val: list[str] = []
        full_name = self._get_index_name_from_name_and_type(name, dtype)
        s_param: Optional[JsonDtoSearchConditionBooleanQueryItem] = self._get_search_parameter(full_name)
        if s_param is not None:
            querry_item_base: JsonDtoSearchConditionBaseQuery = s_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionStrictQuery):
                ret_val.append(querry_item_base.term.value)

            elif isinstance(querry_item_base, JsonDtoSearchConditionBooleanQuery):
                querry_item_first: JsonDtoSearchConditionBaseQuery = querry_item_base.items[0].item
                if isinstance(querry_item_first, JsonDtoSearchConditionStrictQuery):
                    for item in querry_item_base.items:
                        ret_val.append(item.item.term.value)
        return ret_val

    def get_all_search_parameters(self) -> Dict[str, List[str]]:
        """
        Get all search parameters
        :return: A dictionary of search parameters with the name as the key and a list of values as the value
        """
        ret_val: Dict[str, List[str]] = {}
        for search_param in self._search_parameters:
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionStrictQuery):
                name, _ = self._get_facet_name_and_type(querry_item_base.term.field)
                ret_val[name] = [querry_item_base.term.value]
            elif isinstance(querry_item_base, JsonDtoSearchConditionBooleanQuery):
                querry_item_first: JsonDtoSearchConditionBaseQuery = querry_item_base.items[0].item
                if isinstance(querry_item_first, JsonDtoSearchConditionStrictQuery):
                    name, _ = self._get_facet_name_and_type(querry_item_first.term.field)
                    ret_val[name] = []
                    for item in querry_item_base.items:
                        ret_val[name].append(item.item.term.value)
        return ret_val

    def clear_search_parameters(self) -> None:
        """
        Clear all search parameters
        """
        for i in range(len(self._search_parameters) - 1, -1, -1):
            search_param = self._search_parameters[i]
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionStrictQuery):
                self._search_parameters.remove(search_param)
            elif isinstance(querry_item_base, JsonDtoSearchConditionBooleanQuery):
                querry_item_first: JsonDtoSearchConditionBaseQuery = querry_item_base.items[0].item
                if isinstance(querry_item_first, JsonDtoSearchConditionStrictQuery):
                    self._search_parameters.remove(search_param)
        self._counters_up_to_date = False

    def _get_search_parameter(self, full_name: str) -> Optional[JsonDtoSearchConditionBooleanQueryItem]:
        for search_param in self._search_parameters:
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionStrictQuery):
                if querry_item_base.term.field != full_name:
                    continue
                return search_param
            if isinstance(querry_item_base, JsonDtoSearchConditionBooleanQuery):
                querry_item_first: JsonDtoSearchConditionBaseQuery = querry_item_base.items[0].item
                if isinstance(querry_item_first, JsonDtoSearchConditionStrictQuery):
                    if querry_item_first.term.field != full_name:
                        continue
                    return search_param
        return None

    def add_content_search_parameter(self, value: Union[SearchDataType, List[SearchDataType]], remove_old: bool = False) -> bool:
        """
        Add a content search parameter
        :param value: The content search parameter or a list of content search parameters
        :param remove_old: Remove the old search parameters
        :return: If the search parameter was added successfully
        """
        if not isinstance(value, List):
            value = [value]
        return self.add_search_parameter("ContentType", [v.value for v in value], FacedType.NO_TYPE, remove_old)

    def remove_content_search_parameter(self, value: Union[SearchDataType, List[SearchDataType], None]) -> bool:
        """
        Remove a content search parameter
        :param value: The content search parameter or a list of content search parameters to remove, if None all content search parameters are removed
        :return: If the search parameter was removed successfully
        """
        if value is None:
            return self.remove_search_parameter("ContentType", None, FacedType.NO_TYPE)
        if not isinstance(value, List):
            value = [value]
        return self.remove_search_parameter("ContentType", [v.value for v in value], FacedType.NO_TYPE)

    def add_search_parameter_range(
        self, name: str, min_value: float, max_value: float, min_inclusive: bool = True, max_inclusive: bool = True, dtype: FacedType = FacedType.NO_TYPE
    ) -> bool:
        """
        Add a search parameter range
        :param name: The name of the search parameter
        :param min_value: The minimum value
        :param max_value: The maximum value
        :param min_inclusive: Is the minimum value inclusive
        :param max_inclusive: Is the maximum value inclusive
        :param dtype: The data type of the search parameter
        :return: If the search parameter was added
        """
        counter: Optional[JsonFacetedCounter] = None
        for c in self._search_counters:
            if c.faced_name == name and c.faced_type == dtype:
                counter = c
                break
        if counter is None:
            return False
        if not counter.support_range:
            return False
        self.remove_search_parameter_range(name, dtype)
        full_name = self._get_index_name_from_name_and_type(name, dtype)
        s_param = JsonDtoSearchConditionBooleanQueryItem(
            Item=JsonDtoSearchConditionRangeQuery(Field=full_name, Min=min_value, Max=max_value, MinInclusive=min_inclusive, MaxInclusive=max_inclusive),
            Occur=0,
        )
        self._search_parameters.append(s_param)
        self._counters_up_to_date = False
        return True

    def remove_search_parameter_range(self, name: str, dtype: FacedType = FacedType.NO_TYPE) -> bool:
        """
        Remove a search parameter range
        :param name: The name of the search parameter
        :param dtype: The data type of the search parameter
        :return: If the search parameter was removed
        """
        full_name = self._get_index_name_from_name_and_type(name, dtype)
        s_param = self._get_search_parameter_range(full_name)
        if s_param is None:
            return False
        self._search_parameters.remove(s_param)
        self._counters_up_to_date = False
        return True

    def get_search_parameter_range(self, name: str, dtype: FacedType = FacedType.NO_TYPE) -> Optional[Tuple[float, float, bool, bool]]:
        """
        Get the search parameter range
        :param name: The name of the search parameter
        :param dtype: The data type of the search parameter
        :return: The search parameter range or None if it does not exist
        """
        full_name = self._get_index_name_from_name_and_type(name, dtype)
        s_param = self._get_search_parameter_range(full_name)
        if s_param is None:
            return None
        return s_param.item.min, s_param.item.max, s_param.item.min_inclusive, s_param.item.max_inclusive

    def get_all_search_parameters_range(self) -> Dict[str, Tuple[float, float, bool, bool]]:
        """
        Get all search parameters range
        :return: A dictionary of search parameters with the name as the key and a tuple of the min, max, min inclusive and max inclusive as the value
        """
        ret_val: Dict[str, Tuple[float, float, bool, bool]] = {}
        for search_param in self._search_parameters:
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionRangeQuery):
                name, _ = self._get_facet_name_and_type(querry_item_base.field)
                ret_val[name] = (querry_item_base.min, querry_item_base.max, querry_item_base.min_inclusive, querry_item_base.max_inclusive)
        return ret_val

    def clear_search_parameters_range(self) -> None:
        """
        Clear all search parameters range
        :return: None
        """
        for i in range(len(self._search_parameters) - 1, -1, -1):
            search_param = self._search_parameters[i]
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionRangeQuery):
                self._search_parameters.remove(search_param)
        self._counters_up_to_date = False

    def _get_search_parameter_range(self, full_name: str) -> Optional[JsonDtoSearchConditionBooleanQueryItem]:
        self._update_search_names_and_counters()
        for search_param in self._search_parameters:
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionRangeQuery):
                if querry_item_base.field != full_name:
                    continue
                return search_param
        return None

    def add_search_parameter_wildcard(self, value: str) -> bool:
        """
        Add a search parameter wildcard
        :param value: The value of the wildcard
        :return: If the search parameter was added
        """
        self.remove_search_parameter_wildcard()
        s_param = JsonDtoSearchConditionBooleanQueryItem(
            Item=JsonDtoSearchConditionBooleanQuery(
                Items=[
                    JsonDtoSearchConditionBooleanQueryItem(
                        Item=JsonDtoSearchConditionWildcardQuery(Term=JsonDtoSearchConditionTerm(Field="TextC623975962814A5FAAD7FA1CD85DA0DB", Value=value)),
                        Occur=1,
                    ),
                    JsonDtoSearchConditionBooleanQueryItem(
                        Item=JsonDtoSearchConditionWildcardQuery(
                            Term=JsonDtoSearchConditionTerm(Field="DynamicDataC623975962814A5FAAD7FA1CD85DA0DB", Value=value)
                        ),
                        Occur=1,
                    ),
                ]
            ),
            Occur=0,
        )
        self._search_parameters.append(s_param)
        self._counters_up_to_date = False
        return True

    def remove_search_parameter_wildcard(self) -> bool:
        """
        Remove a search parameter wildcard
        :return: If the search parameter was removed
        """
        search_param = self._get_search_parameter_wildcard()
        if search_param is not None:
            self._search_parameters.remove(search_param)
            self._counters_up_to_date = False
            return True
        return False

    def get_search_parameter_wildcard(self) -> Optional[str]:
        """
        Get the search parameter wildcard
        :return: The search parameter wildcard or None if it does not exist
        """
        search_param = self._get_search_parameter_wildcard()
        if search_param is not None:
            return search_param.item.items[0].item.term.value
        return None

    def _get_search_parameter_wildcard(self) -> Optional[JsonDtoSearchConditionBooleanQueryItem]:
        for search_param in self._search_parameters:
            possible_wildcard_name = None
            querry_item_base: JsonDtoSearchConditionBaseQuery = search_param.item
            if isinstance(querry_item_base, JsonDtoSearchConditionBooleanQuery):
                if len(querry_item_base.items) != 2:
                    continue
                for item in querry_item_base.items:
                    if not isinstance(item.item.term, JsonDtoSearchConditionTerm):
                        continue
                    name, dtype = self._get_facet_name_and_type(item.item.term.field)
                    if dtype != FacedType.NO_TYPE:
                        continue
                    if name not in ["Text", "DynamicData"]:
                        continue
                    if possible_wildcard_name is None:
                        possible_wildcard_name = item.item.term.value
                    elif possible_wildcard_name == item.item.term.value:
                        return search_param
        return None

    def get_current_count(self) -> int:
        """
        Get the current count of the search
        :return: The current count of the search
        """
        self._update_search_names_and_counters()
        return self._total_hits

    def get_all_search_names_and_type(self) -> List[Tuple[str, FacedType]]:
        """
        Get all search names and types currently available
        :return: The search names and types currently available
        """
        self._update_search_names_and_counters()
        ret_val: List[Tuple[str, FacedType]] = []
        for counter in self._search_counters:
            ret_val.append((counter.faced_name, counter.faced_type))
        return ret_val

    def get_all_search_names(self) -> List[str]:
        """
        Get all search names currently available
        :return: The search names currently available
        """
        self._update_search_names_and_counters()
        ret_val: List[str] = []
        for counter in self._search_counters:
            ret_val.append(counter.faced_name)
        return ret_val

    def get_all_search_names_range(self) -> List[str]:
        """
        Get all search names that support range
        :return: The search names that support range
        """
        self._update_search_names_and_counters()
        ret_val: List[str] = []
        for counter in self._search_counters:
            if counter.support_range:
                ret_val.append(counter.faced_name)
        return ret_val

    def get_all_search_names_and_type_range(self) -> List[Tuple[str, FacedType]]:
        """
        Get all search names and types that support range
        :return: The search names and types that support range
        """
        self._update_search_names_and_counters()
        ret_val: List[Tuple[str, FacedType]] = []
        for counter in self._search_counters:
            if counter.support_range:
                ret_val.append((counter.faced_name, counter.faced_type))
        return ret_val

    def get_results(self, max_amount: int = 500) -> List[SearchDataBase]:
        """
        Get the search results
        :return: The search results
        """
        max_amount = min(max_amount, 10000)  # Max amount of results per request is 10000
        self._update_search_names_and_counters()
        results: List[SearchDataBase] = []
        while len(results) < max_amount:
            cmd_ret = self._send_command(
                JsonSearchAsyncRequest(
                    Condition=JsonDtoSearchConditionBooleanQuery(Items=self._search_parameters), Limit=min(max_amount - len(results), 10000), Start=len(results)
                ),
                JsonSearchAsyncReturn,
            )

            if not cmd_ret.success:
                raise ConnectionError("Failed to get search results")

            for doc in cmd_ret.documents:
                data = SearchDataBase(altium_workspace=self._altium_workspace)
                for field in doc.fields:
                    name, _ = self._get_facet_name_and_type(field.name)
                    value = field.value

                    for mf_name, mf_type in data.model_fields.items():
                        if mf_type.alias is not None and mf_type.alias == name:
                            name = mf_name
                            break
                    self._add_data_to_data_base(data, name, value)

                results.append(data)
            if len(cmd_ret.documents) < 10000:
                break
        return results

    def _add_data_to_data_base(self, data: SearchDataBase, name: str, value) -> None:
        if not hasattr(data, name):
            data.parameters[name] = value
            return
        if isinstance(getattr(data, name), datetime):
            if value.replace(".", "", 1).isdecimal():
                d = datetime(1899, 12, 31)
                setattr(data, name, d + timedelta(days=float(value)))
            else:
                setattr(data, name, datetime.strptime(value, "%m/%d/%Y %H:%M:%S"))
        elif isinstance(getattr(data, name), int):
            setattr(data, name, int(value))
        elif isinstance(getattr(data, name), bool):
            setattr(data, name, int(value) > 0)
        else:
            setattr(data, name, value)

    def _update_search_names_and_counters(self) -> None:
        if self._counters_up_to_date:
            return
        cmd_ret = self._send_command(
            JsonSearchAsyncRequest(
                Condition=JsonDtoSearchConditionBooleanQuery(Items=self._search_parameters),
            ),
            JsonSearchAsyncReturn,
        )

        if not cmd_ret.success:
            raise ConnectionError("Failed to get search names and counters")

        self._search_counters = cmd_ret.faceted_counters
        for counter in self._search_counters:
            if counter.faced_type == FacedType.NO_TYPE:
                counter.faced_name, counter.faced_type = self._get_facet_name_and_type(counter.faced_name)

        self._search_counters += NOT_COUNTED_SEARCH_PARAMETERS
        self._total_hits = cmd_ret.total
        self._counters_up_to_date = True

    def _get_facet_name_and_type(self, indexed_name: str) -> Tuple[str, FacedType]:
        m1 = re.match("^(.*)_5F([0-9A-F]{8}_[0-9A-F]{6}_[0-9A-F]{6}_[0-9A-F]{6}_[0-9A-F]{46})$", indexed_name)
        m2 = re.match("^(.*)([0-9A-F]{32})$", indexed_name)
        if m1 is not None:
            name = m1.group(1)
            ftype = m1.group(2)
        elif m2 is not None:
            name = m2.group(1)
            ftype = m2.group(2)
        else:
            name = indexed_name
            ftype = ""

        return self._decode_naming_chars(name), FacedType(ftype)

    def _get_index_name_from_name_and_type(self, name: str, ftype: FacedType) -> str:
        if ftype == FacedType.NO_TYPE:
            full_name = self._encode_naming_chars(name)
            full_name += "DD420E8DDD8B445E911A0601BB2B6D53"
        else:
            full_name = self._encode_naming_chars(name)
            full_name += "_5F" + ftype.value
        return full_name

    def _decode_naming_chars(self, inp: str) -> str:
        for key, val in encoding_decoding_naming.items():
            inp = inp.replace(key, val)
        return inp

    def _encode_naming_chars(self, inp: str) -> str:
        for key, val in encoding_decoding_naming.items():
            inp = inp.replace(val, key)
        return inp



from datetime import datetime
from typing import List, Optional

from pydantic_xml import BaseXmlModel, element

from py_altium365.base.connection_handler import ConnectionHandler
from py_altium365.base.enums import PrtGlobalService
from py_altium365.connection.soapy_con import (
    SoapHeader,
    SoapMethod,
    SoapResponse,
    SoapyCon,
)


class SoapUserCredentials(
    BaseXmlModel,
    tag="UserCredentials",
    ns="temp",
    nsmap={"temp": "http://tempuri.org/"},
):
    """SOAP user credentials."""

    user_id: Optional[int] = element(tag="userid", default=None)
    password: str = element(tag="password", nmap={"temp": "http://tempuri.org/"}, ns="temp")


class SoapHeaderCredentials(
    SoapHeader,
    tag="Header",
    ns="soap",
):
    """SOAP header for an API call."""

    user_credentials: SoapUserCredentials


class SoapMethodGetUserWorkspaces(
    SoapMethod,
    tag="GetUserWorkspaces",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP method for login."""


class UserWorkspaceInfo(BaseXmlModel, tag="UserWorkspaceInfo", ns="temp", nsmap={"temp": "http://tempuri.org/"}):
    """User workspace information."""

    workspace_id: int = element(tag="workspaceid")
    owner_id: int = element(tag="ownerid")
    name: str = element(tag="name")
    create_date: datetime = element(tag="createdate")
    status: int = element(tag="status")
    start_date: datetime = element(tag="startdate")
    expiration_date: datetime = element(tag="expirationdate")
    max_user: int = element(tag="maxuser")
    current_user_count: int = element(tag="currentusercount")
    type: int = element(tag="type")
    type_name: str = element(tag="typename")
    hosting_url: str = element(tag="hostingurl")
    space_subscription_guid: str = element(tag="spacesubscriptionguid")
    description: str = element(tag="description", default="")
    display_hosting_url: str = element(tag="displayhostingurl")
    location_id: int = element(tag="locationid")
    creator: str = element(tag="creator")
    location_name: str = element(tag="locationname")
    is_administrator: bool = element(tag="isadministrator", default=False)
    is_default: bool = element(tag="isdefault", default=False)
    status_name: str = element(tag="statusname")
    legacy_hosting_url: str = element(tag="legacyhostingurl")
    owner_account_guid: str = element(tag="owneraccountguid")
    has_administrators: bool = element(tag="hasadministrators", default=False)
    owner_user_guid: str = element(tag="owneruserguid")
    is_secure: bool = element(tag="issecure", default=False)


class SoapGetUserWorkspacesResult(
    BaseXmlModel,
    tag="GetUserWorkspacesResult",
    ns="temp",
    nsmap={"temp": "http://tempuri.org/"},
):
    """SOAP Login result for the SOAP login call."""

    workspaces: List[UserWorkspaceInfo]


class SoapGetUserWorkspacesResponse(
    SoapResponse,
    tag="GetUserWorkspacesResponse",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP method for login."""

    result: SoapGetUserWorkspacesResult


class SoapyConWorkspace(SoapyCon):
    """
    SOAP connection to the Altium workspace.
    """

    def __init__(self, altium_api):
        """
        Initialize the SOAP connection to the Altium portal.
        :param altium_api: The Altium API object.
        """
        workspace_url = altium_api.get_service_url(PrtGlobalService.WORKSPACE)
        if workspace_url is None:
            raise ConnectionError("Failed to get workspace URL")
        super().__init__(ConnectionHandler.get_instance(), workspace_url)
        self._altium_api = altium_api

    def get_user_workspaces(self, session_guid: str):
        """
        Get the user workspaces
        :param session_guid: The session GUID
        :return: The user workspaces
        """
        response = self._send_command(
            SoapHeaderCredentials(user_credentials=SoapUserCredentials(password=session_guid)),
            SoapMethodGetUserWorkspaces(),
            return_method=SoapGetUserWorkspacesResponse,
            soap_action="http://tempuri.org/GetUserWorkspaces",
        )

        return response.result.workspaces



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from pydantic_xml import element, wrapped

from py_altium365.connection.soapy_con import SoapMethod, SoapResponse
from py_altium365.connection.vault.soapy_con_vault_base import (
    AluItem,
    SoapConVaultBase,
    SoapMethodOption,
)

if TYPE_CHECKING:
    from py_altium365.altium_api_workspace import AltiumApiWorkspace  # noqa: F401


class SoapMethodVaultGetAluItems(
    SoapMethod,
    tag="GetALU_Items",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP method for getting ALU items."""

    session_handle: str = element(tag="SessionHandle")
    p_filter: Optional[str] = element(tag="Filter", default=None)
    options: List[SoapMethodOption] = wrapped(
        path="Options",
        entity=element(tag="item"),
        default=[],
    )


class SoapResponseVaultGetAluItems(
    SoapResponse,
    tag="GetALU_ItemsResponse",
    nsmap={"temp": "http://tempuri.org/"},
    ns="temp",
):
    """SOAP response for getting ALU items."""

    records: List[AluItem] = wrapped(
        path="Records",
        tag="item",
        default=[],
    )


class SoapConVault(SoapConVaultBase):
    """SOAP connection class for Altium Vault operations."""

    def get_alu_items(self, p_filter: Optional[str] = None, options: Optional[List[SoapMethodOption]] = None) -> List[AluItem]:
        """
        Get ALU items from the vault.
        :param p_filter: Optional filter string to apply to the query.
        :param options: A list of options to apply to the query.
        :return: The response containing ALU items.
        """
        if options is None:
            options = []
        response = self._send_command(
            header=None,
            method=SoapMethodVaultGetAluItems(session_handle=self._altium_workspace.session_guid, p_filter=p_filter, options=options),
            return_method=SoapResponseVaultGetAluItems,
        )
        return response.records



from typing import Generic, Type, TypeVar

from pydantic import BaseModel
from requests import Session


class JsonRequest(BaseModel):
    """Base class for SOAP method."""


RequestT = TypeVar("RequestT", bound=JsonRequest)


class JsonBase(BaseModel, Generic[RequestT]):
    """Base class for JSON method."""

    request: RequestT


class JsonReturn(BaseModel):
    """Base class for JSON return."""


class JsonCon:
    """Base class for JSON connection."""

    def __init__(self, session: Session, url: str, session_guid: str, host: str):
        """
        Initialize the JsonCon object
        :param session: The main connection session
        :param url: The URL to send the JSON request to
        :param session_guid: The session GUID
        :param host: The host for the host parameter
        """
        self._session: Session = session
        self._url: str = url
        self._session_guid: str = session_guid
        self._host: str = host

    ReturnMethodT = TypeVar("ReturnMethodT", bound=JsonReturn)

    def _send_command(self, request: JsonRequest, return_method: Type[ReturnMethodT]) -> ReturnMethodT:
        headers = {
            "Accept": "application/json",
            "Authorization": f"AFSSessionID {self._session_guid}",
            "host": self._host,
            "content-type": "application/json; charset=utf-8",
            "User-Agent": "Altium Designer",
        }

        json_data = JsonBase(request=request).json(by_alias=True)

        response = self._session.request("REPORT", self._url, data=json_data, headers=headers)

        if response.status_code != 200:
            raise ConnectionError("Failed to send JSON command!")

        return return_method.parse_raw(response.content.decode("utf-8-sig"))



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel
from pydantic_xml import BaseXmlModel, element, wrapped

from py_altium365.base.connection_handler import ConnectionHandler
from py_altium365.connection.soapy_con import SoapMethod, SoapResponse, SoapyCon


class DiscoveryLoginOption(str, Enum):
    """Discovery login options."""

    NONE = "None"
    KILL_EXISTING_SESSION = "KillExistingSession"
    USE_SEPARATE_SESSION = "UseSeparateSession"


class SoapMethodServiceDiscoveryLogin(
    SoapMethod,
    tag="Login",
    nsmap={"altium": "http://altium.com/"},
    ns="altium",
):
    """SOAP method for login to the service discovery."""

    user_name: str = element(tag="userName")
    password: str = element(tag="password")
    secure_login: bool = element(tag="secureLogin", default=False)
    option: DiscoveryLoginOption = element(tag="discoveryLoginOptions", default=DiscoveryLoginOption.NONE)
    product_name: str = element(tag="productName")


class SoapEndPointInfo(BaseXmlModel, tag="EndPointInfo", ns="altium", nsmap={"altium": "http://altium.com/"}):
    """SOAP Parameter."""

    service_kind: str = element(tag="ServiceKind")
    service_url: Optional[str] = element(tag="ServiceUrl", default=None)


class SoapUserParameter(BaseXmlModel, tag="UserParameter", ns="altium", nsmap={"altium": "http://altium.com/"}):
    """SOAP Parameter."""

    name: str = element(tag="Name")
    value: Optional[str] = element(tag="Value", default=None)


class SoapServiceDiscoveryLoginUserInfoResult(
    BaseXmlModel,
    tag="UserInfo",
    ns="altium",
    nsmap={"altium": "http://altium.com/"},
):
    """SOAP Login result for the SOAP login call in the service discovery."""

    session_id: str = element(tag="SessionId")
    user_id: str = element(tag="UserId")
    domain: Optional[str] = element(tag="Domain", default=None)
    account_id: str = element(tag="AccountId")
    email: str = element(tag="Email")
    user_name: str = element(tag="UserName")
    first_name: str = element(tag="FirstName")
    last_name: str = element(tag="LastName")
    full_name: Optional[str] = element(tag="FullName", default=None)
    organisation: str = element(tag="Organisation")
    auth_type: int = element(tag="AuthType")
    parameters: List[SoapUserParameter] = wrapped(
        path="Parameters",
        tag="UserParameter",
        default=[],
    )
    features: List[str] = wrapped(
        path="Features",
        entity=element(tag="string", default=None),
        default=[],
    )


class SoapServiceDiscoveryLoginResult(
    BaseXmlModel,
    tag="LoginResult",
    ns="altium",
    nsmap={"altium": "http://altium.com/"},
):
    """SOAP Login result for the SOAP login call."""

    endpoints: List[SoapEndPointInfo] = wrapped(
        path="Endpoints",
        tag="EndPointInfo",
        default=[],
    )
    user_info: SoapServiceDiscoveryLoginUserInfoResult


class SoapServiceDiscoveryResponse(
    SoapResponse,
    tag="LoginResponse",
    ns="altium",
    nsmap={"altium": "http://altium.com/"},
):
    """SOAP Login response."""

    login_result: SoapServiceDiscoveryLoginResult


class ServiceEndpoints(BaseModel):
    """Service endpoints."""

    ANNOTATIONS: Optional[str] = None
    APPLICATIONS: Optional[str] = None
    BMS: Optional[str] = None
    BOMSERVICE: Optional[str] = None
    BOMSERVICE_AD: Optional[str] = None
    CH: Optional[str] = None
    COMMENTS: Optional[str] = None
    COMMENTSBASE: Optional[str] = None
    COMMENTSUI: Optional[str] = None
    COMPARISONSERVICE: Optional[str] = None
    CollaborationService: Optional[str] = None
    CommentsCloud: Optional[str] = None
    Components: Optional[str] = None
    DDS: Optional[str] = None
    DICTIONARIES: Optional[str] = None
    DSS: Optional[str] = None
    EDS: Optional[str] = None
    EIS: Optional[str] = None
    EXPLORERSERVICE: Optional[str] = None
    FeatureChecking: Optional[str] = None
    IDS: Optional[str] = None
    IDSCloud: Optional[str] = None
    INUSE: Optional[str] = None
    ISR: Optional[str] = None
    Invitation: Optional[str] = None
    LIBRARY_MODELMETADATA_WORKER: Optional[str] = None
    LWTASKS: Optional[str] = None
    Library_Components_Api: Optional[str] = None
    Library_Parts_Api: Optional[str] = None
    MANAGEDFLOWS: Optional[str] = None
    MANAGEDLIBRARIESSERVICE: Optional[str] = None
    MCADCS: Optional[str] = None
    NOTIFICATIONSSERVICE: Optional[str] = None
    PARTCATALOG: Optional[str] = None
    PARTCATALOGUI: Optional[str] = None
    PARTCATALOG_API: Optional[str] = None
    PLMSYNC: Optional[str] = None
    PROJECTCOMPARESERVICE: Optional[str] = None
    PROJECTHISTORYSERVICE: Optional[str] = None
    PROJECTS: Optional[str] = None
    PROJECTSUI: Optional[str] = None
    PUSH: Optional[str] = None
    PushCloud: Optional[str] = None
    REQUIREMENTSSERVICE: Optional[str] = None
    SCHEDULER: Optional[str] = None
    SEARCH: Optional[str] = None
    SEARCHBASE: Optional[str] = None
    SEARCHTEMPLATES: Optional[str] = None
    SECURITY: Optional[str] = None
    SETTINGS: Optional[str] = None
    Sharing: Optional[str] = None
    TASKS: Optional[str] = None
    TC2: Optional[str] = None
    USERSUI: Optional[str] = None
    VAULT: Optional[str] = None
    VAULTUI: Optional[str] = None
    VCSSERVICE: Optional[str] = None
    VIEWER: Optional[str] = None


class SoapyConServiceDiscovery(SoapyCon):
    """SOAP connection to the Altium service discovery."""

    def __init__(self, workspace_url: str):
        super().__init__(ConnectionHandler.get_instance(), workspace_url + "/servicediscovery/servicediscovery.asmx")

        self.service_urls: ServiceEndpoints = ServiceEndpoints()
        self.user_info: Optional[SoapServiceDiscoveryLoginUserInfoResult] = None

    def login(
        self,
        user_name: str,
        password: str,
        secure_login: bool = False,
        option: DiscoveryLoginOption = DiscoveryLoginOption.NONE,
        product_name: str = "Altium Designer",
    ) -> bool:
        """Login to the service discovery."""

        response = self._send_command(
            None,
            SoapMethodServiceDiscoveryLogin(user_name=user_name, password=password, secure_login=secure_login, option=option, product_name=product_name),
            return_method=SoapServiceDiscoveryResponse,
            soap_action="http://altium.com/Login",
        )

        if response.login_result is None:
            return False

        for endpoint in response.login_result.endpoints:
            self.service_urls.__dict__[endpoint.service_kind.replace(".", "_")] = endpoint.service_url
        self.user_info = response.login_result.user_info
        return True
